import json
from os.path import isfile as file_exists


def get_assets(filename='../contracts/contracts.json'):
    if file_exists(filename):
        with open(filename, 'r') as f:
            compiled_interfaces = json.loads(f.read())['contracts']
        # check proper structure e.g.
        # 'Contract1' : { 'abi' : ..., 'bytecode' : ..., 'bytecode_runtime' : ... }, 'Contract2' : ...
        required_members = ['abi', 'bytecode', 'bytecode_runtime']
        for name, interface in compiled_interfaces.items():
            filtered_interface = dict()
            for member in interface.keys():
                # Conversions for unique outputs
                if member == 'abi':
                    filtered_interface['abi'] = interface[member]
                if member == 'bin':
                    filtered_interface['bytecode'] = interface[member]
                if member == 'bin-runtime':
                    filtered_interface['bytecode_runtime'] = interface[member]
            for member in required_members:
                assert member in filtered_interface.keys(), "Contract '{}' doesn't have '{}'!".format(name, member)
            compiled_interfaces[name] = filtered_interface
    else:
        compiled_interfaces = None
    return compiled_interfaces



from web3 import Web3


class Account:
    def __init__(self, w3: Web3, address):
        self._w3 = w3
        self._address = address

    def __repr__(self):
        return self._address

    # Send Ether
    def send(self, address, amount):
        self._w3.eth.sendTransaction({'to': address, 'from': self._address, 'value': amount})

    @property
    def balance(self):
        return self._w3.getBalance(self._address)



from web3 import Web3


class ContractInstance:
    def __init__(self, w3: Web3, address, interface):
        self.__w3 = w3
        self.__address = address
        self.__instance = self.__w3.eth.contract(self.__address, **interface,
                ContractFactoryClass=ImplicitContract)
        # Register new filter to watch for logs from this instance's address
        self.__filter = self.__w3.eth.filter({'address': self.__address})
        self.__logs = self.__w3.utils.filters.LogFilter(self.__filter)

    @property
    def address(self):
        return self.__address

    @property
    def balance(self):
        return self.__w3.eth.getBalance(self.address)

    @property
    def codesize(self):
        return len(self.__w3.eth.getCode(self.address)[2:])/2

    @property
    def hascode(self):
        return self.codesize != 0

    @property
    def logs(self):
        # Returns all the event logs created (since last polled)
        return self.__logs.get()


class ContractFactory:
    def __init__(self, w3: Web3, interface):
        self.__w3 = w3
        self.__interface
        self.__contract_factory = self.__w3.eth.contract(**interface)

    def deploy(self, **kwargs):
        tx_hash = self.__contract_factory.deploy(**kwargs)
        address = self.__w3.eth.getTransactionReceipt(tx_hash)['contractAddress']
        return ContractInstance(self.__w3, address, self.__interface)



import pytest

from web3 import Web3
from web3.contract import ImplicitContract
from web3.providers.eth_tester import EthereumTesterProvider

from eth_tester import EthereumTester
from eth_tester.exceptions import TransactionFailed

from .account import Account
from .contract import ContractFactory
from .asset_loader import get_assets


class Tester:
    def __init__(self):
        self._t = EthereumTester()
        self._w3 = Web3(EthereumTesterProvider(self._t))
        self._compiled_interfaces = get_assets()  # This needs to somehow be configurable
        
        # TODO Set starting balance for all accounts and gas limit for chain
        # NOTE Chain auto-mines transactions vs before

    def contracts(self, name):
        if ':' not in name:
            # If you don't specify which contract in file, use filebase
            name += ':{}'.format(name.split('/')[-1].split('.')[0])
        interface = self._compiled_interfaces[name]
        return self.new_contract(interface)

    def new_contract(self, interface):
        return Contract(self._w3, interface)
        
    @property
    def accounts(self):
        return [Account(self._w3, a) for a in self._t.get_accounts()]
    
    @property
    def tx_fails(self):
        return pytest.raises(TransactionFailed)

    # TODO Potentially replace with TestContract.pastEvents
    def get_log(contract, event_name):
        #TODO Filter by contract and event_name
        #     contract either by address or ABI
        filter_id = t.create_log_filter()
        #TODO Should return decoded event log receipt
        return t.get_only_fiter_changes(filter_id)


@pytest.fixture
def tester():
    return Tester()



import os
from setuptools import setup

# Utility function to read the boilerplate files
def read_md(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name="pytest-ethereum",
    version='0.0.1a1',
    description='py.test plugin for testing Ethereum smart contracts',
    license='MIT',
    long_description=read_md('README.md'),
    author='Bryant Eisenbach',
    author_email='bryant@dappdevs.org',
    url='https://github.com/fubuloubu/pytest-ethereum',
    python_requires='>=3.6',
    packages = ['pytest_ethereum'],
    install_requires=[
        'pytest>=3.4.1',
        'web3>=4.0.0b11',
    ],
    # the following makes a plugin available to pytest
    #entry_points = {
    #    'pytest11': [
    #        'name_of_plugin = pytest_ethereum.pluginmodule',
    #    ]
    #},

    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        'Programming Language :: Python :: 3.6',
        # custom PyPI classifier for pytest plugins
        "Framework :: Pytest",
    ],
)
